@@ -1,42 +1,40 @@
 #ifndef BMSSE2__H__INCLUDED__
 #define BMSSE2__H__INCLUDED__
 /*
-Copyright(c) 2002-2009 Anatoliy Kuznetsov(anatoliy_kuznetsov at yahoo.com)
+Copyright(c) 2002-2017 Anatoliy Kuznetsov(anatoliy_kuznetsov at yahoo.com)
 
-Permission is hereby granted, free of charge, to any person 
-obtaining a copy of this software and associated documentation 
-files (the "Software"), to deal in the Software without restriction, 
-including without limitation the rights to use, copy, modify, merge, 
-publish, distribute, sublicense, and/or sell copies of the Software, 
-and to permit persons to whom the Software is furnished to do so, 
-subject to the following conditions:
+Licensed under the Apache License, Version 2.0 (the "License");
+you may not use this file except in compliance with the License.
+You may obtain a copy of the License at
 
-The above copyright notice and this permission notice shall be included 
-in all copies or substantial portions of the Software.
+    http://www.apache.org/licenses/LICENSE-2.0
 
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
-EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
-OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
-IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
-DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
-ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
-OTHER DEALINGS IN THE SOFTWARE.
-
-For more information please visit:  http://bmagic.sourceforge.net
+Unless required by applicable law or agreed to in writing, software
+distributed under the License is distributed on an "AS IS" BASIS,
+WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+See the License for the specific language governing permissions and
+limitations under the License.
 
+For more information please visit:  http://bitmagic.io
 */
 
+/*! \file bmsse2.h
+    \brief Compute functions for SSE2 SIMD instruction set (internal)
+*/
 
-
-//    Header implements processor specific intrinsics declarations for SSE2
-//    instruction set
 #include<mmintrin.h>
 #include<emmintrin.h>
 
 #include "bmdef.h"
 #include "bmsse_util.h"
+#include "bmutil.h"
 
 
+#ifdef __GNUG__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wconversion"
+#endif
+
 namespace bm
 {
 
@@ -185,54 +183,6 @@ bm::id_t sse2_bit_count_op(const __m128i* BMRESTRICT block,
 }
 
 
-
-
-#define VECT_XOR_ARR_2_MASK(dst, src, src_end, mask)\
-    sse2_xor_arr_2_mask((__m128i*)(dst), (__m128i*)(src), (__m128i*)(src_end), mask)
-
-#define VECT_ANDNOT_ARR_2_MASK(dst, src, src_end, mask)\
-    sse2_andnot_arr_2_mask((__m128i*)(dst), (__m128i*)(src), (__m128i*)(src_end), mask)
-
-#define VECT_BITCOUNT(first, last) \
-    sse2_bit_count((__m128i*) (first), (__m128i*) (last)) 
-
-#define VECT_BITCOUNT_AND(first, last, mask) \
-    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_and) 
-
-#define VECT_BITCOUNT_OR(first, last, mask) \
-    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_or) 
-
-#define VECT_BITCOUNT_XOR(first, last, mask) \
-    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_xor) 
-
-#define VECT_BITCOUNT_SUB(first, last, mask) \
-    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_sub) 
-
-#define VECT_INVERT_ARR(first, last) \
-    sse2_invert_arr(first, last);
-
-#define VECT_AND_ARR(dst, src, src_end) \
-    sse2_and_arr((__m128i*) dst, (__m128i*) (src), (__m128i*) (src_end))
-
-#define VECT_OR_ARR(dst, src, src_end) \
-    sse2_or_arr((__m128i*) dst, (__m128i*) (src), (__m128i*) (src_end))
-
-#define VECT_SUB_ARR(dst, src, src_end) \
-    sse2_sub_arr((__m128i*) dst, (__m128i*) (src), (__m128i*) (src_end))
-
-#define VECT_XOR_ARR(dst, src, src_end) \
-    sse2_xor_arr((__m128i*) dst, (__m128i*) (src), (__m128i*) (src_end))
-
-#define VECT_COPY_BLOCK(dst, src, src_end) \
-    sse2_copy_block((__m128i*) dst, (__m128i*) (src), (__m128i*) (src_end))
-
-#define VECT_SET_BLOCK(dst, dst_end, value) \
-    sse2_set_block((__m128i*) dst, (__m128i*) (dst_end), (value))
-
-
-
-
-
 inline
 bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
                                           const __m128i* BMRESTRICT block_end,
@@ -248,13 +198,13 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
    __m128i m2 = _mm_set_epi32 (mu2, mu2, mu2, mu2);
    __m128i m3 = _mm_set_epi32 (mu3, mu3, mu3, mu3);
    __m128i m4 = _mm_set_epi32 (mu4, mu4, mu4, mu4);
-   __m128i mcnt, ccnt;
+   __m128i mcnt;//, ccnt;
    mcnt = _mm_xor_si128(m1, m1); // bit_cnt = 0
-   ccnt = _mm_xor_si128(m1, m1); // change_cnt = 0
+   //ccnt = _mm_xor_si128(m1, m1); // change_cnt = 0
 
    __m128i tmp1, tmp2;
 
-   int count = (block_end - block)*4; //0;//1;
+   int count = (int)(block_end - block)*4; //0;//1;
 
    bm::word_t  w, w0, w_prev;//, w_l;
    const int w_shift = sizeof(w) * 8 - 1;
@@ -285,6 +235,7 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
        
 
        // compare with zero
+       // SSE4: _mm_test_all_zero()
        {
            // b = (b & 0x55555555) + (b >> 1 & 0x55555555);
            //tmp1 = _mm_srli_epi32(b, 1);                    // tmp1 = (b >> 1 & 0x55555555)
@@ -319,6 +270,8 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
        // ---------------------------------------------------------------------
        {
            //__m128i b = _mm_load_si128(block);
+           // TODO: SSE4...
+           //w = _mm_extract_epi32(b, i);               
 
            const bm::word_t* BMRESTRICT blk = (const bm::word_t*) block;
 
@@ -328,7 +281,7 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
            }
            else
            {
-               if ((w0=blk[0]))
+               if (0!=(w0=blk[0]))
                {
                    BM_INCWORD_BITCOUNT(count, tcnt[0]);
                    count -= !(w_prev ^ (w0 & 1));
@@ -339,7 +292,7 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
                    count -= !w_prev; w_prev ^= w_prev;
                }  
            }
-           if ((w0=blk[1]))
+           if (0!=(w0=blk[1]))
            {
                BM_INCWORD_BITCOUNT(count, tcnt[1]);
                count -= !(w_prev ^ (w0 & 1));
@@ -349,7 +302,7 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
            {
                count -= !w_prev; w_prev ^= w_prev;
            }    
-           if ((w0=blk[2]))
+           if (0!=(w0=blk[2]))
            {
                BM_INCWORD_BITCOUNT(count, tcnt[2]);
                count -= !(w_prev ^ (w0 & 1));
@@ -359,7 +312,7 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
            {
                count -= !w_prev; w_prev ^= w_prev;
            }      
-           if ((w0=blk[3]))
+           if (0!=(w0=blk[3]))
            {
                BM_INCWORD_BITCOUNT(count, tcnt[3]);
                count -= !(w_prev ^ (w0 & 1));
@@ -375,12 +328,146 @@ bm::id_t sse2_bit_block_calc_count_change(const __m128i* BMRESTRICT block,
    _mm_store_si128((__m128i*)tcnt, mcnt);
    *bit_count = tcnt[0] + tcnt[1] + tcnt[2] + tcnt[3];
 
-   return count;
+   return unsigned(count);
 }
 
+#ifdef __GNUG__
+// necessary measure to silence false warning from GCC about negative pointer arithmetics
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Warray-bounds"
+#endif
+
+/*!
+SSE4.2 check for one to two (variable len) 128 bit SSE lines for gap search results (8 elements)
+\internal
+*/
+inline
+unsigned sse2_gap_find(const bm::gap_word_t* BMRESTRICT pbuf, const bm::gap_word_t pos, const unsigned size)
+{
+    BM_ASSERT(size <= 16);
+    BM_ASSERT(size);
+
+    const unsigned unroll_factor = 8;
+    if (size < 4) // for very short vector use conventional scan
+    {
+        unsigned j;
+        for (j = 0; j < size; ++j)
+        {
+            if (pbuf[j] >= pos)
+                break;
+        }
+        return j;
+    }
+
+    __m128i m1, mz, maskF, maskFL;
+
+    mz = _mm_setzero_si128();
+    m1 = _mm_loadu_si128((__m128i*)(pbuf)); // load first 8 elements
+
+    maskF = _mm_cmpeq_epi32(mz, mz); // set all FF
+    maskFL = _mm_slli_si128(maskF, 4 * 2); // byle shift to make [0000 FFFF] 
+    int shiftL = (64 - (unroll_factor - size) * 16);
+    maskFL = _mm_slli_epi64(maskFL, shiftL); // additional bit shift to  [0000 00FF]
+
+    m1 = _mm_andnot_si128(maskFL, m1); // m1 = (~mask) & m1
+    m1 = _mm_or_si128(m1, maskFL);
+
+    __m128i mp = _mm_set1_epi16(pos);  // broadcast pos into all elements of a SIMD vector
+    __m128i  mge_mask = _mm_cmpeq_epi16(_mm_subs_epu16(mp, m1), mz); // unsigned m1 >= mp
+    int mi = _mm_movemask_epi8(mge_mask);  // collect flag bits
+    if (mi)
+    {
+        int bsr_i= bm::bit_scan_fwd(mi) >> 1;
+        return bsr_i;   // address of first one element (target)
+    }
+    if (size == 8)
+        return size;
+
+    // inspect the next lane with possible step back (to avoid over-read the block boundaries)
+    //   GCC gives a false warning for "- unroll_factor" here
+    const bm::gap_word_t* BMRESTRICT pbuf2 = pbuf + size - unroll_factor;
+    BM_ASSERT(pbuf2 > pbuf); // assert in place to make sure GCC warning is indeed false
+
+    m1 = _mm_loadu_si128((__m128i*)(pbuf2)); // load next elements (with possible overlap)
+    mge_mask = _mm_cmpeq_epi16(_mm_subs_epu16(mp, m1), mz); // m1 >= mp        
+    mi = _mm_movemask_epi8(mge_mask);
+    if (mi)
+    {
+        int bsr_i = bm::bit_scan_fwd(mi) >> 1;
+        return size - (unroll_factor - bsr_i);
+    }
+    return size;
+}
+#ifdef __GNUG__
+#pragma GCC diagnostic pop
+#endif
+
+
+#define VECT_XOR_ARR_2_MASK(dst, src, src_end, mask)\
+    sse2_xor_arr_2_mask((__m128i*)(dst), (__m128i*)(src), (__m128i*)(src_end), (bm::word_t)mask)
+
+#define VECT_ANDNOT_ARR_2_MASK(dst, src, src_end, mask)\
+    sse2_andnot_arr_2_mask((__m128i*)(dst), (__m128i*)(src), (__m128i*)(src_end), (bm::word_t)mask)
+
+#define VECT_BITCOUNT(first, last) \
+    sse2_bit_count((__m128i*) (first), (__m128i*) (last))
+
+#define VECT_BITCOUNT_AND(first, last, mask) \
+    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_and)
+
+#define VECT_BITCOUNT_OR(first, last, mask) \
+    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_or)
+
+#define VECT_BITCOUNT_XOR(first, last, mask) \
+    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_xor)
+
+#define VECT_BITCOUNT_SUB(first, last, mask) \
+    sse2_bit_count_op((__m128i*) (first), (__m128i*) (last), (__m128i*) (mask), sse2_sub)
+
+#define VECT_INVERT_BLOCK(first) \
+    sse2_invert_block((__m128i*)first);
+
+#define VECT_AND_BLOCK(dst, src) \
+    sse2_and_block((__m128i*) dst, (__m128i*) (src))
+
+#define VECT_OR_BLOCK(dst, src) \
+    sse2_or_block((__m128i*) dst, (__m128i*) (src))
+
+#define VECT_OR_BLOCK_2WAY(dst, src1, src2) \
+    sse2_or_block_2way((__m128i*) (dst), (__m128i*) (src1), (__m128i*) (src2))
+
+#define VECT_OR_BLOCK_3WAY(dst, src1, src2) \
+    sse2_or_block_3way((__m128i*) (dst), (__m128i*) (src1), (__m128i*) (src2))
+
+#define VECT_OR_BLOCK_5WAY(dst, src1, src2, src3, src4) \
+    sse2_or_block_5way((__m128i*) (dst), (__m128i*) (src1), (__m128i*) (src2), (__m128i*) (src3), (__m128i*) (src4))
+
+#define VECT_SUB_BLOCK(dst, src) \
+    sse2_sub_block((__m128i*) dst, (__m128i*) (src))
+
+#define VECT_XOR_BLOCK(dst, src) \
+    sse2_xor_block((__m128i*) dst, (__m128i*) (src))
+
+#define VECT_XOR_BLOCK_2WAY(dst, src1, src2) \
+    sse2_xor_block_2way((__m128i*) (dst), (const __m128i*) (src1), (const __m128i*) (src2))
+
+#define VECT_COPY_BLOCK(dst, src) \
+    sse2_copy_block((__m128i*) dst, (__m128i*) (src))
+
+#define VECT_STREAM_BLOCK(dst, src) \
+    sse2_stream_block((__m128i*) dst, (__m128i*) (src))
+
+#define VECT_SET_BLOCK(dst, value) \
+    sse2_set_block((__m128i*) dst, value)
+
+
+
 } // namespace
 
 
+#ifdef __GNUG__
+#pragma GCC diagnostic pop
+#endif
 
 
 #endif
