@@ -1,36 +1,34 @@
 #ifndef BMTRANS__H__INCLUDED__
 #define BMTRANS__H__INCLUDED__
-
 /*
-Copyright(c) 2002-2009 Anatoliy Kuznetsov(anatoliy_kuznetsov at yahoo.com)
+Copyright(c) 2002-2017 Anatoliy Kuznetsov(anatoliy_kuznetsov at yahoo.com)
 
-Permission is hereby granted, free of charge, to any person 
-obtaining a copy of this software and associated documentation 
-files (the "Software"), to deal in the Software without restriction, 
-including without limitation the rights to use, copy, modify, merge, 
-publish, distribute, sublicense, and/or sell copies of the Software, 
-and to permit persons to whom the Software is furnished to do so, 
-subject to the following conditions:
+Licensed under the Apache License, Version 2.0 (the "License");
+you may not use this file except in compliance with the License.
+You may obtain a copy of the License at
 
-The above copyright notice and this permission notice shall be included 
-in all copies or substantial portions of the Software.
+    http://www.apache.org/licenses/LICENSE-2.0
 
-THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
-EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
-OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
-IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
-DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
-ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
-OTHER DEALINGS IN THE SOFTWARE.
-
-For more information please visit:  http://bmagic.sourceforge.net
+Unless required by applicable law or agreed to in writing, software
+distributed under the License is distributed on an "AS IS" BASIS,
+WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+See the License for the specific language governing permissions and
+limitations under the License.
 
+For more information please visit:  http://bitmagic.io
 */
+
+/*! \file bmtrans.h
+    \brief Utilities for bit transposition (internal) (experimental!)
+*/
+
 #ifdef _MSC_VER
 #pragma warning( push )
 #pragma warning( disable : 4311 4312 4127)
 #endif
 
+#include "bmdef.h"
+
 namespace bm
 {
 
@@ -43,7 +41,7 @@ struct tmatrix
 {
     typedef T value_type;
     
-    T BM_ALIGN16 value[ROWS][COLS] BM_ALIGN16ATTR;
+    T BM_VECT_ALIGN value[ROWS][COLS] BM_VECT_ALIGN_ATTR;
 
     enum params
     {
@@ -63,8 +61,16 @@ struct tmatrix
     static unsigned rows() { return ROWS; }
     static unsigned cols() { return COLS; }
 
-    const T* row(unsigned row_idx) const { return value[row_idx]; }
-          T* row(unsigned row_idx)       { return value[row_idx]; }
+    const T* row(unsigned row_idx) const
+    {
+        BM_ASSERT(row_idx < ROWS);
+        return value[row_idx];
+    }
+    T* row(unsigned row_idx)
+    {
+        BM_ASSERT(row_idx < ROWS);
+        return value[row_idx];
+    }
 };
 
 
@@ -129,22 +135,22 @@ struct bit_grabber<unsigned short, 16>
     static
     unsigned get(const unsigned short* arr, unsigned j)
     {
-        return  (((arr[0] >> j) & 1) << 0) |
-                (((arr[1] >> j) & 1) << 1) |
-                (((arr[2] >> j) & 1) << 2) |
-                (((arr[3] >> j) & 1) << 3) |
-                (((arr[4] >> j) & 1) << 4) |
-                (((arr[5] >> j) & 1) << 5) |
-                (((arr[6] >> j) & 1) << 6) |
-                (((arr[7] >> j) & 1) << 7) |
-                (((arr[8] >> j) & 1) << 8) |
-                (((arr[9] >> j) & 1) << 9) |
-                (((arr[10]>> j) & 1) << 10)|
-                (((arr[11]>> j) & 1) << 11)|
-                (((arr[12]>> j) & 1) << 12)|
-                (((arr[13]>> j) & 1) << 13)|
-                (((arr[14]>> j) & 1) << 14)|
-                (((arr[15]>> j) & 1) << 15);
+        return  (((arr[0] >> j) & 1u) << 0u) |
+                (((arr[1] >> j) & 1u) << 1u) |
+                (((arr[2] >> j) & 1u) << 2u) |
+                (((arr[3] >> j) & 1u) << 3u) |
+                (((arr[4] >> j) & 1u) << 4u) |
+                (((arr[5] >> j) & 1u) << 5u) |
+                (((arr[6] >> j) & 1u) << 6u) |
+                (((arr[7] >> j) & 1u) << 7u) |
+                (((arr[8] >> j) & 1u) << 8u) |
+                (((arr[9] >> j) & 1u) << 9u) |
+                (((arr[10]>> j) & 1u) << 10u)|
+                (((arr[11]>> j) & 1u) << 11u)|
+                (((arr[12]>> j) & 1u) << 12u)|
+                (((arr[13]>> j) & 1u) << 13u)|
+                (((arr[14]>> j) & 1u) << 14u)|
+                (((arr[15]>> j) & 1u) << 15u);
     }
 };
 
@@ -155,14 +161,15 @@ struct bit_grabber<unsigned char, 8>
     static
     unsigned get(const unsigned char* arr, unsigned j)
     {
-        return  (((arr[0] >> j) & 1) << 0) |
+        return  unsigned(
+                (((arr[0] >> j) & 1) << 0) |
                 (((arr[1] >> j) & 1) << 1) |
                 (((arr[2] >> j) & 1) << 2) |
                 (((arr[3] >> j) & 1) << 3) |
                 (((arr[4] >> j) & 1) << 4) |
                 (((arr[5] >> j) & 1) << 5) |
                 (((arr[6] >> j) & 1) << 6) |
-                (((arr[7] >> j) & 1) << 7);
+                (((arr[7] >> j) & 1) << 7));
     }
 };
 
@@ -233,50 +240,6 @@ struct bit_trans_grabber
     }
 };
 
-/*
-template<>
-struct bit_trans_grabber<unsigned, 32, bm::set_block_plain_size>
-{
-    static
-    unsigned get(const unsigned tmatrix[32][bm::set_block_plain_size], unsigned i, unsigned j)
-    {
-        return
-                (((tmatrix[0][i] >> j) & 1) << 0) |
-                (((tmatrix[1][i] >> j) & 1) << 1) |
-                (((tmatrix[2][i] >> j) & 1) << 2) |
-                (((tmatrix[3][i] >> j) & 1) << 3) |
-                (((tmatrix[4][i] >> j) & 1) << 4) |
-                (((tmatrix[5][i] >> j) & 1) << 5) |
-                (((tmatrix[6][i] >> j) & 1) << 6) |
-                (((tmatrix[7][i] >> j) & 1) << 7) |
-                (((tmatrix[8][i] >> j) & 1) << 8) |
-                (((tmatrix[9][i] >> j) & 1) << 9) |
-                (((tmatrix[10][i] >> j) & 1) << 10) |
-                (((tmatrix[11][i] >> j) & 1) << 11) |
-                (((tmatrix[12][i] >> j) & 1) << 12) |
-                (((tmatrix[13][i] >> j) & 1) << 13) |
-                (((tmatrix[14][i] >> j) & 1) << 14) |
-                (((tmatrix[15][i] >> j) & 1) << 15) |
-                (((tmatrix[16][i] >> j) & 1) << 16) |
-                (((tmatrix[17][i] >> j) & 1) << 17) |
-                (((tmatrix[18][i] >> j) & 1) << 18) |
-                (((tmatrix[19][i] >> j) & 1) << 19) |
-                (((tmatrix[20][i] >> j) & 1) << 20) |
-                (((tmatrix[21][i] >> j) & 1) << 21) |
-                (((tmatrix[22][i] >> j) & 1) << 22) |
-                (((tmatrix[23][i] >> j) & 1) << 23) |
-                (((tmatrix[24][i] >> j) & 1) << 24) |
-                (((tmatrix[25][i] >> j) & 1) << 25) |
-                (((tmatrix[26][i] >> j) & 1) << 26) |
-                (((tmatrix[27][i] >> j) & 1) << 27) |
-                (((tmatrix[28][i] >> j) & 1) << 28) |
-                (((tmatrix[29][i] >> j) & 1) << 29) |
-                (((tmatrix[30][i] >> j) & 1) << 30) |
-                (((tmatrix[31][i] >> j) & 1) << 31);    
-    }
-};
-*/
-
 
 
 /**
@@ -355,14 +318,14 @@ void tmatrix_distance(const T  tmatrix[BPC][BPS],
     for (unsigned i = 0; i < BPC; ++i)
     {
         const T* r1 = tmatrix[i];
-        const T* r1_end = r1 + BPS;
+//        const T* r1_end;// = r1 + BPS;
         distance[i][i] = 
-            bm::bit_block_calc_count((bm::word_t*)r1, (bm::word_t*)r1_end);
+            bm::bit_block_count((bm::word_t*)r1);
 
         for (unsigned j = i + 1; j < BPC; ++j)
         {
             r1 = tmatrix[i];
-            r1_end = r1 + BPS;
+            //r1_end = r1 + BPS;
             unsigned count = 0;
 
             {
@@ -435,6 +398,8 @@ void bit_iblock_make_pcv(
         case total_possible_max:
             pc_vector[i] = ibpc_all_one; 
             continue;
+        default:
+            break;
         }
         
         // Dense-populated set, leave it as is
@@ -501,8 +466,7 @@ void bit_iblock_reduce(
     const unsigned char* BMRESTRICT pc_vector_end,
     unsigned  tmatrix_out[bm::set_block_plain_cnt][bm::set_block_plain_size])
 {
-    ::memset(tmatrix_out, 0, sizeof(tmatrix_out));
-    
+    ::memset(tmatrix_out, 0, bm::set_block_plain_cnt * sizeof(*tmatrix_out));
     unsigned row = 0;
     do 
     {
@@ -607,12 +571,12 @@ void tmatrix_restore(TMatrix& tmatrix,
     typedef typename TMatrix::value_type value_type;
 
     unsigned cols = effective_cols ? effective_cols : tmatrix.cols();
-    for (int row = tmatrix.rows()-1; row >= 0; --row)
+    for (unsigned row = tmatrix.rows()-1u; 1; --row)
     {
-        unsigned ibpc = pc_vector[row] & 7;  
-        int n_row = pc_vector[row] >> 3;
+        unsigned ibpc = pc_vector[row] & 7u;
+        unsigned n_row = pc_vector[row] >> 3u;
 
-        value_type* r1 = tmatrix.row(row);
+        value_type* r1 = tmatrix.row(unsigned(row));
 
         switch(ibpc)
         {
@@ -646,6 +610,9 @@ void tmatrix_restore(TMatrix& tmatrix,
             BM_ASSERT(0);
             break;
         } // switch
+        
+        if (row == 0)
+            break;
     }  // for
 
 }
@@ -656,19 +623,19 @@ void tmatrix_restore(TMatrix& tmatrix,
     \brief Copy GAP block body to bit block with DGap transformation 
     \internal
 */
-template<typename GT>//, typename BT>
+template<typename GT, typename BT>
 void gap_2_bitblock(const GT* BMRESTRICT gap_buf, 
-                          GT* BMRESTRICT block, 
+                          BT* BMRESTRICT block, 
                           unsigned       block_size)
 {
-    GT* dgap_buf = block;
-    GT* block_end = block + block_size;
+    GT* dgap_buf = (GT*) block;
+    BT* block_end = block + block_size;
 
     GT* dgap_end = gap_2_dgap<GT>(gap_buf, dgap_buf, false);
-//    GT* block_end2 = (GT*) block_end;
+    GT* block_end2 = (GT*) block_end;
     
     // zero the tail memory
-    for ( ;dgap_end < block_end; ++dgap_end)
+    for ( ;dgap_end < block_end2; ++dgap_end)
     {
         *dgap_end = 0;
     }
@@ -680,9 +647,11 @@ void gap_2_bitblock(const GT* BMRESTRICT gap_buf,
     @param tmatrix - transposed matrix
     @param pc_vector - row content vector
     @param rstat - output row vector
+    @param effective_cols - effective columns
 
     @internal
 */
+#if 0
 template<class TMatrix>
 void compute_tmatrix_rstat(const TMatrix& tmatrix, 
                            const unsigned char* pc_vector,
@@ -716,8 +685,8 @@ void compute_tmatrix_rstat(const TMatrix& tmatrix,
             bm::bit_count_change32((bm::word_t*)r1, (bm::word_t*)r1_end, 
                                     &rstat[i].bit_count, &rstat[i].gap_count);
 
-            const unsigned bitset_size = sizeof(value_type) * cols;
-            const unsigned total_possible_max_bits = sizeof(value_type)*8*cols;
+            const unsigned bitset_size = unsigned(sizeof(value_type) * cols);
+            const unsigned total_possible_max_bits = unsigned(sizeof(value_type)*8*cols);
 
             rstat[i].best_rep = 
                 bm::best_representation(rstat[i].bit_count,
@@ -734,7 +703,7 @@ void compute_tmatrix_rstat(const TMatrix& tmatrix,
 
     } // for 
 }
-
+#endif
 
 
 /**
@@ -788,18 +757,20 @@ public:
     
     /// Transpose GAP block through a temp. block of aligned(!) memory
     /// 
-    void transpose(const GT* BMRESTRICT gap_buf)
+    void transpose(const GT* BMRESTRICT gap_buf, 
+                         BT* BMRESTRICT tmp_block)
     {
         const unsigned arr_size = BLOCK_SIZE * sizeof(unsigned) / sizeof(GT);
 
         BM_ASSERT(sizeof(tmatrix_.value) == tmatrix_type::n_columns * 
                                             tmatrix_type::n_rows * sizeof(GT));
+                
         // load all GAP as D-GAP(but not head word) into aligned bit-block
-        gap_2_bitblock(gap_buf, tmp_gap_block_, BLOCK_SIZE * 2);
+        gap_2_bitblock(gap_buf, tmp_block, BLOCK_SIZE);
         
         // transpose
         vect_bit_transpose<GT, tmatrix_type::n_rows, tmatrix_type::n_columns>
-                           (tmp_gap_block_, arr_size, tmatrix_.value);
+                           ((GT*)tmp_block, arr_size, tmatrix_.value);
 
         // calculate number of non-zero columns
         eff_cols_ = find_effective_columns(tmatrix_);        
@@ -808,19 +779,20 @@ public:
 	/// Transpose array of shorts
 	///
 	void transpose(const GT* BMRESTRICT garr,
-		           unsigned garr_size)
+		           unsigned garr_size,
+				   BT* BMRESTRICT tmp_block)
 	{
 		BM_ASSERT(garr_size);
 
-		bit_block_set(tmp_gap_block_, 0);
-		::memcpy(tmp_gap_block_, garr, sizeof(GT)*garr_size);
+		bit_block_set(tmp_block, 0);
+		::memcpy(tmp_block, garr, sizeof(GT)*garr_size);
 
         const unsigned arr_size = BLOCK_SIZE * sizeof(unsigned) / sizeof(GT);
         BM_ASSERT(sizeof(tmatrix_.value) == tmatrix_type::n_columns * 
                                             tmatrix_type::n_rows * sizeof(GT));
         // transpose
         vect_bit_transpose<GT, tmatrix_type::n_rows, tmatrix_type::n_columns>
-                           (tmp_gap_block_, arr_size, tmatrix_.value);
+                           ((GT*)tmp_block, arr_size, tmatrix_.value);
 
         // calculate number of non-zero columns
         eff_cols_ = find_effective_columns(tmatrix_);        
@@ -846,7 +818,7 @@ public:
     void reduce()
     {
         tmatrix_reduce(tmatrix_, pc_vector_, eff_cols_);
-        compute_tmatrix_rstat(tmatrix_, pc_vector_, rstat_vector_, eff_cols_);
+        //compute_tmatrix_rstat(tmatrix_, pc_vector_, rstat_vector_, eff_cols_);
     }
 
     void restore()
@@ -858,18 +830,20 @@ public:
     /// Restore GAP block from the transposed matrix
     ///
     void trestore(GT             gap_head, 
-                  GT* BMRESTRICT gap_buf)
+                  GT* BMRESTRICT gap_buf, 
+                  BT* BMRESTRICT tmp_block)
     {
         BM_ASSERT(sizeof(tmatrix_.value) == tmatrix_type::n_columns * 
                                             tmatrix_type::n_rows * sizeof(GT));
   
         // restore into a temp buffer
-        GT* gap_tmp = tmp_gap_block_;
+        GT* gap_tmp = (GT*)tmp_block;
+        //*gap_tmp++ = gap_head;
        
         vect_bit_trestore<GT, tmatrix_type::n_rows, tmatrix_type::n_columns>(tmatrix_.value, gap_tmp);
         
         // D-Gap to GAP block recalculation
-        gap_tmp = tmp_gap_block_;
+        gap_tmp = (GT*)tmp_block;
         dgap_2_gap<GT>(gap_tmp, gap_buf, gap_head);
     }
     
@@ -881,13 +855,12 @@ public:
     unsigned char                 pc_vector_[tmatrix_type::n_rows];
     unsigned                      pc_vector_stat_[bm::ibpc_end];
     typename tmatrix_type::rstat  rstat_vector_[tmatrix_type::n_rows];
-
-    GT  BM_ALIGN16                tmp_gap_block_[BLOCK_SIZE*2] BM_ALIGN16ATTR;
 };
 
 
 } // namespace bm
 
+
 #ifdef _MSC_VER
 #pragma warning( pop )
 #endif
